from __future__ import print_function
import time
from collections import OrderedDict
from datetime import datetime, timedelta
from threading import _Timer as Timer
from functools import wraps
from dateutil.relativedelta import relativedelta
from dateutil.rrule import rrule
from dateutil.parser import parse
from . import YEARLY, MONTHLY, DAILY, HOURLY, MINUTELY, SECONDLY
from . import LogStub, Settings
from . import NEW, WAITING, COMPLETE, DONE, FAILED, RECURRING, DELAYED


class Threader(Timer):
    def __init__(self, job):
        self.job = job
        super(self.__class__, self).__init__(self.job.seconds_to_run, self.job.function,
                         self.job.args, self.job.kwargs)

    def run(self):
        if self.interval >= 0:
            self.job.status = WAITING
            try:
                self.finished.wait(self.interval)
                self.job.log.info('%s is now scheduled to run at %s' % (
                    self.job.name, datetime.now() + timedelta(seconds=self.interval)))
            except OverflowError:
                while self.interval > Settings.sleep_max:
                    self.interval -= Settings.sleep_max
                    self.finished.wait(Settings.sleep_max)
                self.finished.wait(self.interval)
            if not self.finished.is_set():
                try:
                    result = self.function(*self.args, **self.kwargs)
                    if result and result is not self.job.function:
                        self.job.results[datetime.now()] = result
                    self.finished.set()
                    self.job.status = COMPLETE
                    self.job.run_count += 1
                    self.job.error_count = 0
                except Exception as e:
                    self.job.status = FAILED
                    self.job.error_count += 1
                    self.job.log.error('Error in', self.job.name, 'error count at',
                                       self.job.error_count, 'Traceback:', e)
                    raise


class Job(object):
    run_count = 0
    error_count = 0
    seconds_to_run = None
    next_run_time = None
    thread = None

    def __init__(self, func, **settings):
        self.function = func
        self.id = id(func)
        self.name = settings.get('name', func.__name__)
        self.args = settings.get('args', [])
        self.kwargs = settings.get('kwargs', {})
        self.job_type = settings.get('job_type')
        self.status = settings.get('status', Settings.status)
        self.retrys = settings.get('retrys', Settings.retrys)
        self.retry_delay = settings.get('retry_delay', Settings.retry_delay)
        self.how_many = settings.get('how_many', Settings.how_many)
        self.log = settings.get('log', LogStub())
        self.results = OrderedDict()


class RecuringJob(Job):
    def __init__(self, func, **settings):
        super(self.__class__, self).__init__(func,
                         name=settings.get('name'),
                         args=settings.get('args', []),
                         kwargs=settings.get('kwargs', {}),
                         job_type=RECURRING,
                         status=settings.get('status', Settings.status),
                         retrys=settings.get('retrys', Settings.retrys),
                         retry_delay=settings.get('retry_delay', Settings.retry_delay),
                         how_many=settings.get('how_many', Settings.how_many))
        self.interval = settings.get('interval', DAILY)
        self.days_of_week = settings.get('days_of_week')
        self.start_datetime = settings.get('start_datetime')

    def find_next_time(self):
        time_delta = self.start_datetime - datetime.now()
        if time_delta.days < 0:
            if self.interval == SECONDLY:
                self.start_datetime = datetime.now()
            elif self.interval == MINUTELY:
                self.start_datetime = self.start_datetime.replace(
                    datetime.now().year,
                    datetime.now().month,
                    datetime.now().day,
                    datetime.now().hour,
                    datetime.now().minute,
                    microsecond=0)
            elif self.interval == HOURLY:
                self.start_datetime = self.start_datetime.replace(
                    datetime.now().year,
                    datetime.now().month,
                    datetime.now().day,
                    datetime.now().hour,
                    microsecond=0)
            elif self.interval == DAILY:
                self.start_datetime = self.start_datetime.replace(
                    datetime.now().year,
                    datetime.now().month,
                    datetime.now().day,
                    microsecond=0)
            elif self.interval == MONTHLY:
                self.start_datetime = self.start_datetime.replace(
                    datetime.now().year,
                    datetime.now().month,
                    microsecond=0)
            elif self.interval == YEARLY:
                self.start_datetime = self.start_datetime.replace(
                    datetime.now().year,
                    microsecond=0)
        runs = list(rrule(self.interval, dtstart=self.start_datetime, count=2))
        self.seconds_to_run = (runs[0] - datetime.now()).total_seconds()
        self.next_run_time = runs[1]
        count = 1
        while True:
            if self.seconds_to_run > 0:
                break
            runs = list(rrule(self.interval, dtstart=self.start_datetime, count=(count + 2)))
            self.seconds_to_run = (runs[count] - datetime.now()).total_seconds()
            self.next_run_time = runs[count + 1]
            count += 1

    def schedule_next(self):
        if self.error_count < self.retrys:
            if self.run_count < self.how_many:
                if self.status is not WAITING:
                    if self.status is NEW:
                        self.find_next_time()
                    elif self.status is FAILED:
                        self.seconds_to_run = self.retry_delay
                    else:
                        self.seconds_to_run = (self.next_run_time - datetime.now()).total_seconds()
                    if self.seconds_to_run > 0:
                        self.thread = Threader(self)
                        self.thread.start()
                    if self.status is not NEW:
                        self.find_next_time()
            else:
                self.status = DONE
        else:
            self.status = DONE


class DelayedJob(Job):
    def __init__(self, func, **settings):
        super(self.__class__, self).__init__(func,
                         name=settings.get('name'),
                         args=settings.get('args', []),
                         kwargs=settings.get('kwargs', {}),
                         job_type=DELAYED,
                         status=settings.get('status', Settings.status),
                         retrys=settings.get('retrys', Settings.retrys),
                         retry_delay=settings.get('retry_delay', Settings.retry_delay),
                         how_many=settings.get('how_many', Settings.how_many))
        self.time_delta = settings.get('time_delta')
        self.years = settings.get('years')
        self.months = settings.get('months')
        self.days = settings.get('days')
        self.hours = settings.get('hours')
        self.minutes = settings.get('minutes')
        self.seconds = settings.get('seconds')

    def schedule_next(self):
        if self.error_count < self.retrys:
            if self.run_count < self.how_many:
                if self.status is not WAITING:
                    if self.status is FAILED:
                        self.seconds_to_run = self.retry_delay
                    else:
                        self.time_delta = datetime.now() + relativedelta(
                            years=self.years,
                            months=self.months,
                            days=self.days,
                            hours=self.hours,
                            minutes=self.minutes,
                            seconds=self.seconds,
                            microseconds=0) - datetime.now()
                        # The line above may seem strange but it creates a
                        # timedelta object; datetime + relativedelta = new_datetime
                        # new_datetime - datetime = timedelta
                        self.seconds_to_run = self.time_delta.total_seconds()
                    self.thread = Threader(self)
                    self.thread.start()
            else:
                self.status = DONE
        else:
            self.status = DONE


class Scheduler:
    job_queue = OrderedDict()

    def __init__(self):
        def event_loop():
            while True:
                start = time.clock()
                for key in self.job_queue:
                    if self.job_queue[key].status is DONE:
                        del self.job_queue[key]
                    elif self.job_queue[key].status is not WAITING:
                        self.job_queue[key].schedule_next()
                wait = Settings.loop_wait - (time.clock() - start)
                if wait >= 0:
                    time.sleep(wait)

        Timer(0, event_loop).start()

    def schedule_recuring_job(self,
                              days_of_week=None,
                              how_many=Settings.how_many,
                              interval=DAILY,
                              log=None,
                              name=None,
                              retry_delay=Settings.retry_delay,
                              retrys=Settings.retrys,
                              start_date=Settings.start_date,
                              start_time=Settings.start_time):
        def wrap(f):
            def wrapped(*args, **kwargs):
                self.add_job(RecuringJob(f,
                                         args=args or [],
                                         days_of_week=days_of_week,
                                         how_many=how_many,
                                         interval=interval,
                                         kwargs=kwargs or {},
                                         log=log,
                                         name=name or self.func_name(f),
                                         retrys=retrys,
                                         retry_delay=retry_delay,
                                         start_datetime=parse('%s %s' % (start_date, start_time),
                                                              yearfirst=True,
                                                              ignoretz=True)))

            return wraps(f)(wrapped)

        return wrap

    def add_recuring_job(self, f,
                         args=None,
                         days_of_week=None,
                         how_many=Settings.how_many,
                         interval=DAILY,
                         kwargs=None,
                         log=None,
                         name=None,
                         retry_delay=Settings.retry_delay,
                         retrys=Settings.retrys,
                         start_date=Settings.start_date,
                         start_time=Settings.start_time):
        self.add_job(RecuringJob(f,
                                 args=args or [],
                                 days_of_week=days_of_week,
                                 how_many=how_many,
                                 interval=interval,
                                 kwargs=kwargs or {},
                                 log=log,
                                 name=name or self.func_name(f),
                                 retry_delay=retry_delay,
                                 retrys=retrys,
                                 start_datetime=parse('%s %s' % (start_date, start_time),
                                                      yearfirst=True,
                                                      ignoretz=True)))

    def schedule_delayed_job(self,
                             days=0,
                             hours=0,
                             how_many=Settings.how_many,
                             log=None,
                             minutes=0,
                             months=0,
                             name=None,
                             retrys=Settings.retrys,
                             retry_delay=Settings.retry_delay,
                             seconds=0,
                             years=0):
        def wrap(f):
            def wrapped(*args, **kwargs):
                self.add_job(DelayedJob(f,
                                        args=args or [],
                                        days=days,
                                        hours=hours,
                                        how_many=how_many,
                                        kwargs=kwargs or {},
                                        log=log,
                                        minutes=minutes,
                                        months=months,
                                        name=name or self.func_name(f),
                                        retry_delay=retry_delay,
                                        retrys=retrys,
                                        seconds=seconds,
                                        years=years))

            return wraps(f)(wrapped)

        return wrap

    def add_delayed_job(self, f,
                        args=None,
                        days=0,
                        hours=0,
                        how_many=Settings.how_many,
                        kwargs=None,
                        log=None,
                        minutes=0,
                        months=0,
                        name=None,
                        retrys=Settings.retrys,
                        retry_delay=Settings.retry_delay,
                        seconds=0,
                        years=0):
        self.add_job(DelayedJob(f,
                                args=args or [],
                                days=days,
                                hours=hours,
                                how_many=how_many,
                                kwargs=kwargs or {},
                                log=log,
                                minutes=minutes,
                                months=months,
                                name=name or self.func_name(f),
                                retry_delay=retry_delay,
                                retrys=retrys,
                                seconds=seconds,
                                years=years))

    def func_name(self, func):
        if type(func) is not str:
            try:
                func = func.__name__
            except AttributeError:
                raise NameError('%s does not seem to be a valid function or name.' % func)
        return func

    def add_job(self, job):
        self.job_queue[job.name or self.func_name(job.function)] = job

    def remove_job(self, func):
        func = self.func_name(func)
        try:
            try:
                self.job_queue[func].thread.finished.set()
            except AttributeError:
                pass
            del self.job_queue[func]
            print(func, 'was removed.')
        except KeyError:
            print('The job', func, 'was not in the Queue')

    def get_results(self, func, run_date=None, run_time=None):
        func = self.func_name(func)
        return self.job_queue[func].results

    # Is primarily being used for debuging
    def get_job(self, func):
        func = self.func_name(func)
        try:
            return self.job_queue[func]
        except KeyError:
            print('The job', func, 'was not in the Queue')